@@ -1,10 +1,8 @@
-# This Source Code Form is subject to the terms of the Mozilla Public
-# License, v. 2.0. If a copy of the MPL was not distributed with this
-# file, You can obtain one at http://mozilla.org/MPL/2.0/.
-#
-# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
+# SPDX-FileCopyrightText: 2022 James R. Barlow
+# SPDX-License-Identifier: MPL-2.0
+
+"""Implement some features in Python and monkey-patch them onto C++ classes.
 
-"""
 In several cases the implementation of some higher levels features might as
 well be in Python. Fortunately we can attach Python methods to C++ class
 bindings after the fact.
@@ -12,40 +10,29 @@ bindings after the fact.
 We can also move the implementation to C++ if desired.
 """
 
+from __future__ import annotations
+
 import datetime
-import inspect
 import mimetypes
-import platform
 import shutil
 from collections.abc import KeysView, MutableMapping
 from decimal import Decimal
 from io import BytesIO
 from pathlib import Path
-from subprocess import PIPE, run
+from subprocess import run
 from tempfile import NamedTemporaryFile
-from typing import (
-    Any,
-    BinaryIO,
-    Callable,
-    ItemsView,
-    Iterator,
-    List,
-    Optional,
-    Tuple,
-    Type,
-    TypeVar,
-    Union,
-    ValuesView,
-)
+from typing import BinaryIO, Callable, ItemsView, Iterator, TypeVar, ValuesView
 from warnings import warn
 
 from . import Array, Dictionary, Name, Object, Page, Pdf, Stream
+from ._augments import augment_override_cpp, augments
 from ._qpdf import (
     AccessMode,
     AttachedFile,
     AttachedFileSpec,
     Attachments,
     NameTree,
+    NumberTree,
     ObjectStreamMode,
     Rectangle,
     StreamDecodeLevel,
@@ -64,127 +51,6 @@ __all__ = []
 Numeric = TypeVar('Numeric', int, float, Decimal)
 
 
-def augment_override_cpp(fn):
-    """This Python function should replace the C++ implementation, if there is one."""
-    fn._augment_override_cpp = True
-    return fn
-
-
-def augment_if_no_cpp(fn):
-    """This Python function provides a Python implementation if no C++ implementation exists."""
-    fn._augment_if_no_cpp = True
-    return fn
-
-
-def _is_inherited_method(meth):
-    # Augmenting a C++ with a method that cls inherits from the Python
-    # object is never what we want.
-    return meth.__qualname__.startswith('object.')
-
-
-def _is_augmentable(m):
-    return (
-        inspect.isfunction(m) and not _is_inherited_method(m)
-    ) or inspect.isdatadescriptor(m)
-
-
-def augments(cls_cpp: Type[Any]):
-    """Attach methods of a Python support class to an existing class
-
-    This monkeypatches all methods defined in the support class onto an
-    existing class. Example:
-
-    .. code-block:: python
-
-        @augments(ClassDefinedInCpp)
-        class SupportClass:
-            def foo(self):
-                pass
-
-    The Python method 'foo' will be monkeypatched on ClassDefinedInCpp. SupportClass
-    has no meaning on its own and should not be used, but gets returned from
-    this function so IDE code inspection doesn't get too confused.
-
-    We don't subclass because it's much more convenient to monkeypatch Python
-    methods onto the existing Python binding of the C++ class. For one thing,
-    this allows the implementation to be moved from Python to C++ or vice
-    versa. It saves having to implement an intermediate Python subclass and then
-    ensures that the C++ superclass never 'leaks' to pikepdf users. Finally,
-    wrapper classes and subclasses can become problematic if the call stack
-    crosses the C++/Python boundary multiple times.
-
-    Any existing methods may be used, regardless of whether they are defined
-    elsewhere in the support class or in the target class.
-
-    For data fields to work, the target class must be
-    tagged ``py::dynamic_attr`` in pybind11.
-
-    Strictly, the target class does not have to be C++ or derived from pybind11.
-    This works on pure Python classes too.
-
-    THIS DOES NOT work for class methods.
-
-    (Alternative ideas: https://github.com/pybind/pybind11/issues/1074)
-    """
-
-    OVERRIDE_WHITELIST = {'__eq__', '__hash__', '__repr__'}
-    if platform.python_implementation() == 'PyPy':
-        # Either PyPy or pybind11's interface to PyPy automatically adds a __getattr__
-        OVERRIDE_WHITELIST |= {'__getattr__'}  # pragma: no cover
-
-    def class_augment(cls, cls_cpp=cls_cpp):
-
-        # inspect.getmembers has different behavior on PyPy - in particular it seems
-        # that a typical PyPy class like cls will have more methods that it considers
-        # methods than CPython does. Our predicate should take care of this.
-        for name, member in inspect.getmembers(cls, predicate=_is_augmentable):
-            if name == '__weakref__':
-                continue
-            if (
-                hasattr(cls_cpp, name)
-                and hasattr(cls, name)
-                and name not in getattr(cls, '__abstractmethods__', set())
-                and name not in OVERRIDE_WHITELIST
-                and not getattr(getattr(cls, name), '_augment_override_cpp', False)
-            ):
-                if getattr(getattr(cls, name), '_augment_if_no_cpp', False):
-                    # If tagged as "augment if no C++", we are supported older pybind11
-                    # versions that may not define some functions. If pybind11 provides
-                    # the required function, don't override it.
-                    continue
-
-                # If the original C++ class and Python support class both define the
-                # same name, we generally have a conflict, because this is augmentation
-                # not inheritance. However, if the method provided by the support class
-                # is an abstract method, then we can consider the C++ version the
-                # implementation. Also, pybind11 provides defaults for __eq__,
-                # __hash__ and __repr__ that we often do want to override directly.
-
-                raise RuntimeError(  # pragma: no cover
-                    f"C++ {cls_cpp} and Python {cls} both define the same "
-                    f"non-abstract method {name}: "
-                    f"{getattr(cls_cpp, name, '')!r}, "
-                    f"{getattr(cls, name, '')!r}"
-                )
-            if inspect.isfunction(member):
-                setattr(cls_cpp, name, member)
-                installed_member = getattr(cls_cpp, name)
-                installed_member.__qualname__ = member.__qualname__.replace(
-                    cls.__name__, cls_cpp.__name__
-                )
-            elif inspect.isdatadescriptor(member):
-                setattr(cls_cpp, name, member)
-
-        def disable_init(self):
-            # Prevent initialization of the support class
-            raise NotImplementedError(self.__class__.__name__ + '.__init__')
-
-        cls.__init__ = disable_init
-        return cls
-
-    return class_augment
-
-
 def _single_page_pdf(page) -> bytes:
     """Construct a single page PDF from the provided page in memory."""
     pdf = Pdf.new()
@@ -265,10 +131,10 @@ class Extend_Object:
         for k in del_keys:
             del self[k]  # pylint: disable=unsupported-delete-operation
 
-    def _type_check_write(self, filter, decode_parms):
-        if isinstance(filter, list):
-            filter = Array(filter)
-        filter = filter.wrap_in_array()
+    def _type_check_write(self, filter_, decode_parms):
+        if isinstance(filter_, list):
+            filter_ = Array(filter_)
+        filter_ = filter_.wrap_in_array()
 
         if isinstance(decode_parms, list):
             decode_parms = Array(decode_parms)
@@ -277,7 +143,7 @@ class Extend_Object:
         else:
             decode_parms = decode_parms.wrap_in_array()
 
-        if not all(isinstance(item, Name) for item in filter):
+        if not all(isinstance(item, Name) for item in filter_):
             raise TypeError(
                 "filter must be: pikepdf.Name or pikepdf.Array([pikepdf.Name])"
             )
@@ -288,32 +154,34 @@ class Extend_Object:
                 "decode_parms must be: pikepdf.Dictionary or "
                 "pikepdf.Array([pikepdf.Dictionary])"
             )
-        if len(decode_parms) != 0 and len(filter) != len(decode_parms):
+        if len(decode_parms) != 0 and len(filter_) != len(decode_parms):
             raise ValueError(
-                f"filter ({repr(filter)}) and decode_parms "
+                f"filter ({repr(filter_)}) and decode_parms "
                 f"({repr(decode_parms)}) must be arrays of same length"
             )
-        if len(filter) == 1:
-            filter = filter[0]
+        if len(filter_) == 1:
+            filter_ = filter_[0]
         if len(decode_parms) == 0:
             decode_parms = None
         elif len(decode_parms) == 1:
             decode_parms = decode_parms[0]
-        return filter, decode_parms
+        return filter_, decode_parms
 
     def write(
         self,
         data: bytes,
         *,
-        filter: Union[Name, Array, None] = None,
-        decode_parms: Union[Dictionary, Array, None] = None,
+        filter: Name | Array | None = None,
+        decode_parms: Dictionary | Array | None = None,
         type_check: bool = True,
     ):  # pylint: disable=redefined-builtin
         """
         Replace stream object's data with new (possibly compressed) `data`.
 
-        `filter` and `decode_parms` specify that compression that is present on
-        the input `data`.
+        `filter` and `decode_parms` describe any compression that is already
+        present on the input `data`. For example, if your data is already
+        compressed with the Deflate algorithm, you would set
+        ``filter=Name.FlateDecode``.
 
         When writing the PDF in :meth:`pikepdf.Pdf.save`,
         pikepdf may change the compression or apply compression to data that was
@@ -341,7 +209,6 @@ class Extend_Object:
         `decode_parms` is an Array of Dictionary, where each array index
         is corresponds to the filter.
         """
-
         if type_check and filter is not None:
             filter, decode_parms = self._type_check_write(filter, decode_parms)
 
@@ -354,11 +221,10 @@ class Extend_Pdf:
         self, include=None, exclude=None
     ):  # pylint: disable=unused-argument
         """
-        Present options to IPython or Jupyter for rich display of this object
+        Present options to IPython or Jupyter for rich display of this object.
 
         See https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
         """
-
         bio = BytesIO()
         self.save(bio)
         bio.seek(0)
@@ -506,11 +372,13 @@ class Extend_Pdf:
         return Stream(self, data, d, **kwargs)
 
     def add_blank_page(
-        self, *, page_size: Tuple[Numeric, Numeric] = (612.0, 792.0)
+        self, *, page_size: tuple[Numeric, Numeric] = (612.0, 792.0)
     ) -> Page:
         """
-        Add a blank page to this PDF. If pages already exist, the page will be added to
-        the end. Pages may be reordered using ``Pdf.pages``.
+        Add a blank page to this PDF.
+
+        If pages already exist, the page will be added to the end. Pages may be
+        reordered using ``Pdf.pages``.
 
         The caller may add content to the page by modifying its objects after creating
         it.
@@ -590,9 +458,11 @@ class Extend_Pdf:
         """
         return EncryptionInfo(self._encryption_data)
 
-    def check(self) -> List[str]:
+    def check(self) -> list[str]:
         """
-        Check if PDF is well-formed.  Similar to ``qpdf --check``.
+        Check if PDF is well-formed.
+
+        Similar to ``qpdf --check``.
         """
 
         class DiscardingParser(StreamParser):
@@ -605,7 +475,7 @@ class Extend_Pdf:
             def handle_eof(self):
                 pass
 
-        problems: List[str] = []
+        problems: list[str] = []
 
         self._decode_all_streams_and_discard()
 
@@ -620,22 +490,23 @@ class Extend_Pdf:
 
     def save(
         self,
-        filename_or_stream: Union[Path, str, BinaryIO, None] = None,
+        filename_or_stream: Path | str | BinaryIO | None = None,
         *,
         static_id: bool = False,
         preserve_pdfa: bool = True,
-        min_version: Union[str, Tuple[str, int]] = "",
-        force_version: Union[str, Tuple[str, int]] = "",
+        min_version: str | tuple[str, int] = "",
+        force_version: str | tuple[str, int] = "",
         fix_metadata_version: bool = True,
         compress_streams: bool = True,
-        stream_decode_level: Optional[StreamDecodeLevel] = None,
+        stream_decode_level: StreamDecodeLevel | None = None,
         object_stream_mode: ObjectStreamMode = ObjectStreamMode.preserve,
         normalize_content: bool = False,
         linearize: bool = False,
         qdf: bool = False,
         progress: Callable[[int], None] = None,
-        encryption: Optional[Union[Encryption, bool]] = None,
+        encryption: Encryption | bool | None = None,
         recompress_flate: bool = False,
+        deterministic_id: bool = False,
     ) -> None:
         """
         Save all modifications to this :class:`pikepdf.Pdf`.
@@ -652,8 +523,10 @@ class Extend_Pdf:
 
             static_id: Indicates that the ``/ID`` metadata, normally
                 calculated as a hash of certain PDF contents and metadata
-                including the current time, should instead be generated
-                deterministically. Normally for debugging.
+                including the current time, should instead be set to a static
+                value. Only use this for debugging and testing. Use
+                ``deterministic_id`` if you want to get the same ``/ID`` for
+                the same document contents.
             preserve_pdfa: Ensures that the file is generated in a
                 manner compliant with PDF/A and other stricter variants.
                 This should be True, the default, in most cases.
@@ -727,6 +600,14 @@ class Extend_Pdf:
                 Alternately, an ``Encryption`` object may be provided that
                 sets the parameters for new encryption.
 
+            deterministic_id: Indicates that the ``/ID`` metadata, normally
+                calculated as a hash of certain PDF contents and metadata
+                including the current time, should instead be computed using
+                only deterministic data like the file contents. At a small
+                runtime cost, this enables generation of the same ``/ID`` if
+                the same inputs are converted in the same way multiple times.
+                Does not work for encrypted files.
+
         Raises:
             PdfError
             ForeignObjectError
@@ -786,13 +667,14 @@ class Extend_Pdf:
             encryption=encryption,
             samefile_check=getattr(self, '_tmp_stream', None) is None,
             recompress_flate=recompress_flate,
+            deterministic_id=deterministic_id,
         )
 
     @staticmethod
     def open(
-        filename_or_stream: Union[Path, str, BinaryIO],
+        filename_or_stream: Path | str | BinaryIO,
         *,
-        password: Union[str, bytes] = "",
+        password: str | bytes = "",
         hex_password: bool = False,
         ignore_xref_streams: bool = False,
         suppress_warnings: bool = True,
@@ -832,7 +714,6 @@ class Extend_Pdf:
         release that object when appropriate.
 
         Examples:
-
             >>> with Pdf.open("test.pdf") as pdf:
                     ...
 
@@ -873,6 +754,7 @@ class Extend_Pdf:
                 entire input file into memory at open time; this will use more
                 memory and may recent performance especially when the opened
                 file will not be modified.
+
         Raises:
             pikepdf.PasswordError: If the password failed to open the
                 file.
@@ -941,14 +823,6 @@ class Extend_ObjectMapping:
         except KeyError:
             return default
 
-    @augment_if_no_cpp  # Remove this shim when dropping pybind11 < 2.8.0 support
-    def keys(self):
-        return KeysView(self)
-
-    @augment_if_no_cpp  # Remove this shim when dropping pybind11 < 2.8.0 support
-    def values(self):
-        return (v for _k, v in self.items())
-
 
 def check_is_box(obj) -> None:
     try:
@@ -971,7 +845,7 @@ def check_is_box(obj) -> None:
 class Extend_Page:
     @property
     def mediabox(self):
-        "This page's /MediaBox, in PDF units."
+        """Return page's /MediaBox, in PDF units."""
         return self._get_mediabox(True)
 
     @mediabox.setter
@@ -981,11 +855,11 @@ class Extend_Page:
 
     @property
     def cropbox(self):
-        """This page's effective /CropBox, in PDF units.
+        """Return page's effective /CropBox, in PDF units.
 
         If the /CropBox is not defined, the /MediaBox is returned.
         """
-        return self._get_cropbox(True)
+        return self._get_cropbox(True, False)
 
     @cropbox.setter
     def cropbox(self, value):
@@ -994,12 +868,12 @@ class Extend_Page:
 
     @property
     def trimbox(self):
-        """This page's effective /TrimBox, in PDF units.
+        """Return page's effective /TrimBox, in PDF units.
 
         If the /TrimBox is not defined, the /CropBox is returned (and if
         /CropBox is not defined, /MediaBox is returned).
         """
-        return self._get_trimbox(True)
+        return self._get_trimbox(True, False)
 
     @trimbox.setter
     def trimbox(self, value):
@@ -1008,7 +882,11 @@ class Extend_Page:
 
     @property
     def images(self) -> _ObjectMapping:
-        """Return all images associated with this page."""
+        """Return all regular images associated with this page.
+
+        This method does not recurse into Form XObjects and does not
+        attempt to find inline images.
+        """
         return self._images
 
     @property
@@ -1020,12 +898,12 @@ class Extend_Page:
         self,
         res: Object,
         res_type: Name,
-        name: Optional[Name] = None,
+        name: Name | None = None,
         *,
         prefix: str = '',
         replace_existing: bool = True,
     ) -> Name:
-        """Adds a new resource to the page's Resources dictionary.
+        """Add a new resource to the page's Resources dictionary.
 
         If the Resources dictionaries do not exist, they will be created.
 
@@ -1042,8 +920,6 @@ class Extend_Page:
                 Mutually exclusive with name parameter.
             replace_existing: If the name already exists in one of the resource
                 dictionaries, remove it.
-        Returns:
-            The name of the object.
 
         Example:
             >>> resource_name = pdf.pages[0].add_resource(formxobj, Name.XObject)
@@ -1054,6 +930,10 @@ class Extend_Page:
             If *res* does not belong to the same `Pdf` that owns this page,
             a copy of *res* is automatically created and added instead. In previous
             versions, it was necessary to change for this case manually.
+
+        .. versionchanged:: 4.3.0
+            Returns the name of the overlay in the resources dictionary instead
+            of returning None.
         """
         if Name.Resources not in self.obj:
             self.obj.Resources = Dictionary()
@@ -1082,15 +962,19 @@ class Extend_Page:
         return name
 
     def _over_underlay(
-        self, other, rect: Optional[Rectangle], under: bool, push_stack: bool
-    ) -> None:
+        self,
+        other,
+        rect: Rectangle | None,
+        under: bool,
+        push_stack: bool,
+        shrink: bool,
+        expand: bool,
+    ) -> Name:
         formx = None
         if isinstance(other, Page):
-            page = other
             formx = other.as_form_xobject()
         elif isinstance(other, Dictionary) and other.get(Name.Type) == Name.Page:
-            page = Page(other)
-            formx = page.as_form_xobject()
+            formx = Page(other).as_form_xobject()
         elif (
             isinstance(other, Stream)
             and other.get(Name.Type) == Name.XObject
@@ -1099,13 +983,17 @@ class Extend_Page:
             formx = other
 
         if formx is None:
-            raise TypeError("other object is not something we can convert to FormX")
+            raise TypeError(
+                "other object is not something we can convert to Form XObject"
+            )
 
         if rect is None:
-            rect = Rectangle(page.trimbox)
+            rect = Rectangle(self.trimbox)
 
         formx_placed_name = self.add_resource(formx, Name.XObject)
-        cs = self.calc_form_xobject_placement(formx, formx_placed_name, rect)
+        cs = self.calc_form_xobject_placement(
+            formx, formx_placed_name, rect, allow_shrink=shrink, allow_expand=expand
+        )
 
         if push_stack:
             self.contents_add(b'q\n', prepend=True)  # prepend q
@@ -1113,14 +1001,17 @@ class Extend_Page:
 
         self.contents_add(cs, prepend=under)
         self.contents_coalesce()
+        return formx_placed_name
 
     def add_overlay(
         self,
-        other: Union[Object, Page],
-        rect: Optional[Rectangle] = None,
+        other: Object | Page,
+        rect: Rectangle | None = None,
         *,
-        push_stack: Optional[bool] = True,
-    ):
+        push_stack: bool = True,
+        shrink: bool = True,
+        expand: bool = True,
+    ) -> Name:
         """Overlay another object on this page.
 
         Overlays will be drawn after all previous content, potentially drawing on top
@@ -1130,40 +1021,88 @@ class Extend_Page:
             other: A Page or Form XObject to render as an overlay on top of this
                 page.
             rect: The PDF rectangle (in PDF units) in which to draw the overlay.
-                If omitted, this page's trimbox, cropbox or mediabox will be used.
+                If omitted, this page's trimbox, cropbox or mediabox (in that order)
+                will be used.
             push_stack: If True (default), push the graphics stack of the existing
                 content stream to ensure that the overlay is rendered correctly.
                 Officially PDF limits the graphics stack depth to 32. Most
                 viewers will tolerate more, but excessive pushes may cause problems.
-                Multiple content streams may also be coalseced into a single content
+                Multiple content streams may also be coalesced into a single content
                 stream where this parameter is True, since the PDF specification
                 permits PDF writers to coalesce streams as they see fit.
+            shrink: If True (default), allow the object to shrink to fit inside the
+                rectangle. The aspect ratio will be preserved.
+            expand: If True (default), allow the object to expand to fit inside the
+                rectangle. The aspect ratio will be preserved.
+
+        Returns:
+            The name of the Form XObject that contains the overlay.
 
         .. versionadded:: 2.14
 
-        .. versionchanged:: 3.3.0
+        .. versionchanged:: 4.0.0
+            Added the *push_stack* parameter. Previously, this method behaved
+            as if *push_stack* were False.
+
+        .. versionchanged:: 4.2.0
+            Added the *shrink* and *expand* parameters. Previously, this method
+            behaved as if ``shrink=True, expand=False``.
+
+        .. versionchanged:: 4.3.0
+            Returns the name of the overlay in the resources dictionary instead
+            of returning None.
         """
-        return self._over_underlay(other, rect, under=False, push_stack=push_stack)
+        return self._over_underlay(
+            other,
+            rect,
+            under=False,
+            push_stack=push_stack,
+            expand=expand,
+            shrink=shrink,
+        )
 
     def add_underlay(
-        self, other: Union[Object, Page], rect: Optional[Rectangle] = None
-    ):
+        self,
+        other: Object | Page,
+        rect: Rectangle | None = None,
+        *,
+        shrink: bool = True,
+        expand: bool = True,
+    ) -> Name:
         """Underlay another object beneath this page.
 
         Underlays will be drawn before all other content, so they may be overdrawn
         partially or completely.
 
+        There is no *push_stack* parameter for this function, since adding an
+        underlay can be done without manipulating the graphics stack.
+
         Args:
             other: A Page or Form XObject to render as an underlay underneath this
                 page.
             rect: The PDF rectangle (in PDF units) in which to draw the underlay.
-                If omitted, this page's MediaBox will be used.
+                If omitted, this page's trimbox, cropbox or mediabox (in that order)
+                will be used.
+            shrink: If True (default), allow the object to shrink to fit inside the
+                rectangle. The aspect ratio will be preserved.
+            expand: If True (default), allow the object to expand to fit inside the
+                rectangle. The aspect ratio will be preserved.
+
+        Returns:
+            The name of the Form XObject that contains the underlay.
 
         .. versionadded:: 2.14
-        """
-        return self._over_underlay(other, rect, under=True, push_stack=False)
 
-    def contents_add(self, contents: Union[Stream, bytes], *, prepend: bool = False):
+        .. versionchanged:: 4.2.0
+            Added the *shrink* and *expand* parameters. Previously, this method
+            behaved as if ``shrink=True, expand=False``. Fixed issue with wrong
+            page rect being selected.
+        """
+        return self._over_underlay(
+            other, rect, under=True, push_stack=False, expand=expand, shrink=shrink
+        )
+
+    def contents_add(self, contents: Stream | bytes, *, prepend: bool = False):
         """Append or prepend to an existing page's content stream.
 
         Args:
@@ -1180,14 +1119,16 @@ class Extend_Page:
     @augment_override_cpp
     def __setattr__(self, name, value):
         if hasattr(self.__class__, name):
-            return object.__setattr__(self, name, value)
-        setattr(self.obj, name, value)
+            object.__setattr__(self, name, value)
+        else:
+            setattr(self.obj, name, value)
 
     @augment_override_cpp
     def __delattr__(self, name):
         if hasattr(self.__class__, name):
-            return object.__delattr__(self, name)
-        delattr(self.obj, name)
+            object.__delattr__(self, name)
+        else:
+            delattr(self.obj, name)
 
     def __getitem__(self, key):
         return self.obj[key]
@@ -1201,9 +1142,6 @@ class Extend_Page:
     def __contains__(self, key):
         return key in self.obj
 
-    def __eq__(self, other):
-        return self.obj == other.obj
-
     def get(self, key, default=None):
         try:
             return self[key]
@@ -1282,12 +1220,15 @@ class Extend_Attachments(MutableMapping):
 @augments(AttachedFileSpec)
 class Extend_AttachedFileSpec:
     @staticmethod
-    def from_filepath(pdf: Pdf, path: Union[Path, str], *, description: str = ''):
+    def from_filepath(pdf: Pdf, path: Path | str, *, description: str = ''):
         """Construct a file specification from a file path.
 
         This function will automatically add a creation and modified date
         using the file system, and a MIME type inferred from the file's extension.
 
+        If the data required for the attach is in memory, use
+        :meth:`pikepdf.AttachedFileSpec` instead.
+
         Args:
             pdf: The Pdf to attach this file specification to.
             path: A file path for the file to attach to this Pdf.
@@ -1319,14 +1260,13 @@ class Extend_AttachedFileSpec:
                 f"<pikepdf._qpdf.AttachedFileSpec for {self.filename!r}, "
                 f"description {self.description!r}>"
             )
-        else:
-            return f"<pikepdf._qpdf.AttachedFileSpec description {self.description!r}>"
+        return f"<pikepdf._qpdf.AttachedFileSpec description {self.description!r}>"
 
 
 @augments(AttachedFile)
 class Extend_AttachedFile:
     @property
-    def creation_date(self) -> Optional[datetime.datetime]:
+    def creation_date(self) -> datetime.datetime | None:
         if not self._creation_date:
             return None
         return decode_pdf_date(self._creation_date)
@@ -1336,7 +1276,7 @@ class Extend_AttachedFile:
         self._creation_date = encode_pdf_date(value)
 
     @property
-    def mod_date(self) -> Optional[datetime.datetime]:
+    def mod_date(self) -> datetime.datetime | None:
         if not self._mod_date:
             return None
         return decode_pdf_date(self._mod_date)
@@ -1357,14 +1297,7 @@ class Extend_AttachedFile:
 
 
 @augments(NameTree)
-class Extend_NameTree(MutableMapping):
-    def __len__(self):
-        return len(self._as_map())
-
-    def __iter__(self):
-        for name, _value in self._nameval_iter():
-            yield name
-
+class Extend_NameTree:
     def keys(self):
         return KeysView(self._as_map())
 
@@ -1374,27 +1307,34 @@ class Extend_NameTree(MutableMapping):
     def items(self):
         return ItemsView(self._as_map())
 
-    def __eq__(self, other):
-        return self.obj.objgen == other.obj.objgen
+    get = MutableMapping.get
+    pop = MutableMapping.pop
+    popitem = MutableMapping.popitem
+    clear = MutableMapping.clear
+    update = MutableMapping.update
+    setdefault = MutableMapping.setdefault
 
-    def __contains__(self, name: Union[str, bytes]) -> bool:
-        """
-        Returns True if the name tree contains the specified name.
 
-        Args:
-            name (str or bytes): The name to search for in the name tree.
-                This is not a PDF /Name object, but an arbitrary key.
-                If name is a *str*, we search the name tree for the UTF-8
-                encoded form of name. If *bytes*, we search for a key
-                equal to those bytes.
-        """
-        return self._contains(name)
+MutableMapping.register(NameTree)
 
-    def __getitem__(self, name: Union[str, bytes]) -> Object:
-        return self._getitem(name)
 
-    def __setitem__(self, name: Union[str, bytes], o: Object):
-        self._setitem(name, o)
+@augments(NumberTree)
+class Extend_NumberTree:
+    def keys(self):
+        return KeysView(self._as_map())
 
-    def __delitem__(self, name: Union[str, bytes]):
-        self._delitem(name)
+    def values(self):
+        return ValuesView(self._as_map())
+
+    def items(self):
+        return ItemsView(self._as_map())
+
+    get = MutableMapping.get
+    pop = MutableMapping.pop
+    popitem = MutableMapping.popitem
+    clear = MutableMapping.clear
+    update = MutableMapping.update
+    setdefault = MutableMapping.setdefault
+
+
+MutableMapping.register(NumberTree)
