@@ -1,20 +1,20 @@
-# This Source Code Form is subject to the terms of the Mozilla Public
-# License, v. 2.0. If a copy of the MPL was not distributed with this
-# file, You can obtain one at http://mozilla.org/MPL/2.0/.
-#
-# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
+# SPDX-FileCopyrightText: 2022 James R. Barlow
+# SPDX-License-Identifier: MPL-2.0
+
+"""Extract images embedded in PDF."""
+
+from __future__ import annotations
 
-import struct
 from abc import ABC, abstractmethod
 from decimal import Decimal
 from io import BytesIO
 from itertools import zip_longest
 from pathlib import Path
 from shutil import copyfileobj
-from typing import NamedTuple, Optional
+from typing import Any, BinaryIO, Callable, NamedTuple, Sequence, TypeVar, cast
 
-from PIL import Image, ImageCms
-from PIL.TiffTags import TAGS_V2 as TIFF_TAGS
+from PIL import Image
+from PIL.ImageCms import ImageCmsProfile
 
 from pikepdf import (
     Array,
@@ -28,60 +28,65 @@ from pikepdf import (
     String,
     jbig2,
 )
+from pikepdf._exceptions import DependencyError
+from pikepdf._qpdf import Buffer
+from pikepdf._version import __version__
+from pikepdf.models import _transcoding
 
-
-class DependencyError(Exception):
-    "A third party dependency is needed to extract images of this type."
+T = TypeVar('T')
 
 
 class UnsupportedImageTypeError(Exception):
-    "This image is formatted in a way pikepdf does not supported."
+    """This image is formatted in a way pikepdf does not supported."""
 
 
 class NotExtractableError(Exception):
-    "Indicates that an image cannot be directly extracted."
+    """Indicates that an image cannot be directly extracted."""
 
 
 class HifiPrintImageNotTranscodableError(NotExtractableError):
-    "Image contains high fidelity printing information and cannot be extracted."
+    """Image contains high fidelity printing information and cannot be extracted."""
 
 
 class InvalidPdfImageError(Exception):
-    "This image is not valid according to the PDF 1.7 specification."
+    """This image is not valid according to the PDF 1.7 specification."""
 
 
-def array_str(value):
+def _array_str(value: Object | str | list):
     """Simplify pikepdf objects to array of str. Keep Streams and dictionaries intact."""
 
-    def _array_str(item):
+    def _convert(item):
         if isinstance(item, (list, Array)):
-            return [_array_str(subitem) for subitem in item]
-        elif isinstance(item, (Stream, Dictionary, bytes, int)):
+            return [_convert(subitem) for subitem in item]
+        if isinstance(item, (Stream, Dictionary, bytes, int)):
             return item
-        elif isinstance(item, (Name, str)):
+        if isinstance(item, (Name, str)):
             return str(item)
-        elif isinstance(item, (String)):
+        if isinstance(item, (String)):
             return bytes(item)
-        else:
-            raise NotImplementedError(value)
+        raise NotImplementedError(value)
 
-    result = _array_str(value)
+    result = _convert(value)
     if not isinstance(result, list):
         result = [result]
     return result
 
 
-def dict_or_array_dict(value):
+def _ensure_list(value: list[Object] | Dictionary | Array) -> list[Object]:
+    """Ensure value is a list of pikepdf.Object, if it was not already.
+
+    To support DecodeParms which can be present as either an array of dicts or a single
+    dict. It's easier to convert to an array of one dict.
+    """
     if isinstance(value, list):
         return value
-    if isinstance(value, Dictionary):
-        return [value.as_dict()]
-    if isinstance(value, Array):
-        return [v for v in value.as_list()]
-    raise NotImplementedError(value)
+    return list(value.wrap_in_array().as_list())
 
 
-def metadata_from_obj(obj, name, type_, default):
+def _metadata_from_obj(
+    obj: Dictionary | Stream, name: str, type_: Callable[[Any], T], default: T
+) -> T | None:
+    """Retrieve metadata from a dictionary or stream, and ensure it is the expected type."""
     val = getattr(obj, name, default)
     try:
         return type_(val)
@@ -105,52 +110,53 @@ class PaletteData(NamedTuple):
 
 
 class PdfImageBase(ABC):
+    """Abstract base class for images."""
 
     SIMPLE_COLORSPACES = {'/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray'}
     MAIN_COLORSPACES = SIMPLE_COLORSPACES | {'/DeviceCMYK', '/CalCMYK', '/ICCBased'}
     PRINT_COLORSPACES = {'/Separation', '/DeviceN'}
 
     @abstractmethod
-    def _metadata(self, name, type_, default):
+    def _metadata(self, name: str, type_: Callable[[Any], T], default: T) -> T:
         """Get metadata for this image type."""
 
     @property
-    def width(self):
+    def width(self) -> int:
         """Width of the image data in pixels."""
-        return self._metadata('Width', int, None)
+        return self._metadata('Width', int, 0)
 
     @property
-    def height(self):
+    def height(self) -> int:
         """Height of the image data in pixels."""
-        return self._metadata('Height', int, None)
+        return self._metadata('Height', int, 0)
 
     @property
-    def image_mask(self):
-        """``True`` if this is an image mask."""
+    def image_mask(self) -> bool:
+        """Return ``True`` if this is an image mask."""
         return self._metadata('ImageMask', bool, False)
 
     @property
-    def _bpc(self):
+    def _bpc(self) -> int | None:
         """Bits per component for this image (low-level)."""
-        return self._metadata('BitsPerComponent', int, None)
+        return self._metadata('BitsPerComponent', int, 0)
 
     @property
     def _colorspaces(self):
         """Colorspace (low-level)."""
-        return self._metadata('ColorSpace', array_str, [])
+        return self._metadata('ColorSpace', _array_str, [])
 
     @property
     def filters(self):
         """List of names of the filters that we applied to encode this image."""
-        return self._metadata('Filter', array_str, [])
+        return self._metadata('Filter', _array_str, [])
 
     @property
     def decode_parms(self):
         """List of the /DecodeParms, arguments to filters."""
-        return self._metadata('DecodeParms', dict_or_array_dict, [])
+        return self._metadata('DecodeParms', _ensure_list, [])
 
     @property
-    def colorspace(self) -> Optional[str]:
+    def colorspace(self) -> str | None:
         """PDF name of the colorspace that best describes this image."""
         if self.image_mask:
             return None  # Undefined for image masks
@@ -174,25 +180,20 @@ class PdfImageBase(ABC):
         )
 
     @property
-    def bits_per_component(self):
+    def bits_per_component(self) -> int:
         """Bits per component of this image."""
-        if self._bpc is None:
+        if self._bpc is None or self._bpc == 0:
             return 1 if self.image_mask else 8
         return self._bpc
 
     @property
     @abstractmethod
-    def is_inline(self) -> bool:
-        """Is this an inline image?"""
+    def icc(self) -> ImageCmsProfile | None:
+        """Return ICC profile for this image if one is defined."""
 
     @property
-    @abstractmethod
-    def icc(self) -> Image.Image:
-        """Returns ICC profile for this image if one is defined."""
-
-    @property
-    def indexed(self):
-        """``True`` if the image has a defined color palette."""
+    def indexed(self) -> bool:
+        """Check if the image has a defined color palette."""
         return '/Indexed' in self._colorspaces
 
     def _colorspace_has_name(self, name):
@@ -207,17 +208,17 @@ class PdfImageBase(ABC):
         return False
 
     @property
-    def is_device_n(self):
-        """``True`` if image has a /DeviceN (complex printing) colorspace."""
+    def is_device_n(self) -> bool:
+        """Check if image has a /DeviceN (complex printing) colorspace."""
         return self._colorspace_has_name('/DeviceN')
 
     @property
-    def is_separation(self):
-        """``True`` if image has a /DeviceN (complex printing) colorspace."""
+    def is_separation(self) -> bool:
+        """Check if image has a /DeviceN (complex printing) colorspace."""
         return self._colorspace_has_name('/Separation')
 
     @property
-    def size(self):
+    def size(self) -> tuple[int, int]:
         """Size of image as (width, height)."""
         return self.width, self.height
 
@@ -237,13 +238,12 @@ class PdfImageBase(ABC):
         return mode_from_xcolor_space.get(xcolor_space, '')
 
     @property
-    def mode(self):
-        """``PIL.Image.mode`` equivalent for this image, where possible
+    def mode(self) -> str:
+        """``PIL.Image.mode`` equivalent for this image, where possible.
 
         If an ICC profile is attached to the image, we still attempt to resolve a Pillow
         mode.
         """
-
         m = ''
         if self.is_device_n:
             m = 'DeviceN'
@@ -251,22 +251,21 @@ class PdfImageBase(ABC):
             m = 'Separation'
         elif self.indexed:
             m = 'P'
-        elif self.bits_per_component == 1:
+        elif self.colorspace == '/DeviceGray' and self.bits_per_component == 1:
             m = '1'
-        elif self.bits_per_component == 8:
-            if self.colorspace == '/DeviceRGB':
-                m = 'RGB'
-            elif self.colorspace == '/DeviceGray':
-                m = 'L'
-            elif self.colorspace == '/DeviceCMYK':
-                m = 'CMYK'
-            elif self.colorspace == '/ICCBased':
-                try:
-                    m = self._approx_mode_from_icc()
-                except (ValueError, TypeError) as e:
-                    raise NotImplementedError(
-                        "Not sure how to handle PDF image of this type"
-                    ) from e
+        elif self.colorspace == '/DeviceGray' and self.bits_per_component > 1:
+            m = 'L'
+        elif self.colorspace == '/DeviceRGB':
+            m = 'RGB'
+        elif self.colorspace == '/DeviceCMYK':
+            m = 'CMYK'
+        elif self.colorspace == '/ICCBased':
+            try:
+                m = self._approx_mode_from_icc()
+            except (ValueError, TypeError) as e:
+                raise NotImplementedError(
+                    "Not sure how to handle PDF image of this type"
+                ) from e
         if m == '':
             raise NotImplementedError(
                 "Not sure how to handle PDF image of this type"
@@ -275,7 +274,9 @@ class PdfImageBase(ABC):
 
     @property
     def filter_decodeparms(self):
-        """PDF has a lot of optional data structures concerning /Filter and
+        """Return normalized the Filter and DecodeParms data.
+
+        PDF has a lot of possible data structures concerning /Filter and
         /DecodeParms. /Filter can be absent or a name or an array, /DecodeParms
         can be absent or a dictionary (if /Filter is a name) or an array (if
         /Filter is an array). When both are arrays the lengths match.
@@ -288,9 +289,8 @@ class PdfImageBase(ABC):
         return list(zip_longest(self.filters, self.decode_parms, fillvalue={}))
 
     @property
-    def palette(self) -> Optional[PaletteData]:
-        """Retrieves the color palette for this image if applicable."""
-
+    def palette(self) -> PaletteData | None:
+        """Retrieve the color palette for this image if applicable."""
         if not self.indexed:
             return None
         try:
@@ -323,14 +323,13 @@ class PdfImageBase(ABC):
         """Convert this PDF image to a Python PIL (Pillow) image."""
 
     @staticmethod
-    def _remove_simple_filters(obj, filters):
+    def _remove_simple_filters(obj: Stream, filters: Sequence[str]):
         """Remove simple lossless compression where it appears.
 
         Args:
-            obj (pikepdf.Stream): the compressed object
-            filters (list of str): all files on the data
+            obj: the compressed object
+            filters: all files on the data
         """
-
         COMPLEX_FILTERS = {
             '/DCTDecode',
             '/JPXDecode',
@@ -365,7 +364,7 @@ class PdfImageBase(ABC):
 
 
 class PdfImage(PdfImageBase):
-    """Support class to provide a consistent API for manipulating PDF images
+    """Support class to provide a consistent API for manipulating PDF images.
 
     The data structure for images inside PDFs is irregular and complex,
     making it difficult to use without introducing errors for less
@@ -374,7 +373,11 @@ class PdfImage(PdfImageBase):
     Pillow imaging library.
     """
 
+    obj: Stream
+    _icc: ImageCmsProfile | None
+
     def __new__(cls, obj):
+        """Construct a PdfImage... or a PdfJpxImage if that is what we really are."""
         instance = super().__new__(cls)
         instance.__init__(obj)
         if '/JPXDecode' in instance.filters:
@@ -382,14 +385,13 @@ class PdfImage(PdfImageBase):
             instance.__init__(obj)
         return instance
 
-    def __init__(self, obj):
-        """Construct a PDF image from a Image XObject inside a PDF
+    def __init__(self, obj: Stream):
+        """Construct a PDF image from a Image XObject inside a PDF.
 
         ``pim = PdfImage(page.Resources.XObject['/ImageNN'])``
 
         Args:
-            obj (pikepdf.Object): an Image XObject
-
+            obj: an Image XObject
         """
         if isinstance(obj, Stream) and obj.stream_dict.get("/Subtype") != "/Image":
             raise TypeError("can't construct PdfImage from non-image")
@@ -403,7 +405,7 @@ class PdfImage(PdfImageBase):
 
     @classmethod
     def _from_pil_image(cls, *, pdf, page, name, image):  # pragma: no cover
-        """Insert a PIL image into a PDF (rudimentary)
+        """Insert a PIL image into a PDF (rudimentary).
 
         Args:
             pdf (pikepdf.Pdf): the PDF to attach the image to
@@ -411,7 +413,6 @@ class PdfImage(PdfImageBase):
             name (str or pikepdf.Name): the name to set the image
             image (PIL.Image.Image): the image to insert
         """
-
         data = image.tobytes()
 
         imstream = Stream(pdf, data)
@@ -430,12 +431,7 @@ class PdfImage(PdfImageBase):
         return cls(imstream)
 
     def _metadata(self, name, type_, default):
-        return metadata_from_obj(self.obj, name, type_, default)
-
-    @property
-    def is_inline(self):
-        """``False`` for image XObject."""
-        return False
+        return _metadata_from_obj(self.obj, name, type_, default)
 
     @property
     def _iccstream(self):
@@ -447,13 +443,10 @@ class PdfImage(PdfImageBase):
         raise NotImplementedError("Don't know how to find ICC stream for image")
 
     @property
-    def icc(self):
+    def icc(self) -> ImageCmsProfile | None:
         """If an ICC profile is attached, return a Pillow object that describe it.
 
         Most of the information may be found in ``icc.profile``.
-
-        Returns:
-            PIL.ImageCms.ImageCmsProfile
         """
         if self.colorspace not in ('/ICCBased', '/Indexed'):
             return None
@@ -462,7 +455,7 @@ class PdfImage(PdfImageBase):
             iccbuffer = iccstream.get_stream_buffer()
             iccbytesio = BytesIO(iccbuffer)
             try:
-                self._icc = ImageCms.ImageCmsProfile(iccbytesio)
+                self._icc = ImageCmsProfile(iccbytesio)
             except OSError as e:
                 if str(e) == 'cannot open profile from string':
                     # ICC profile is corrupt
@@ -471,32 +464,40 @@ class PdfImage(PdfImageBase):
                     ) from e
         return self._icc
 
-    def _extract_direct(self, *, stream) -> str:
-        """Attempt to extract the image directly to a usable image file
+    def _extract_direct(self, *, stream: BinaryIO) -> str:
+        """Attempt to extract the image directly to a usable image file.
 
         If there is no way to extract the image without decompressing or
         transcoding then raise an exception. The type and format of image
         generated will vary.
 
         Args:
-            stream: Writable stream to write data to
+            stream: Writable file stream to write data to, e.g. an open file
         """
 
-        def normal_dct_rgb():
+        def normal_dct_rgb() -> bool:
             # Normal DCTDecode RGB images have the default value of
             # /ColorTransform 1 and are actually in YUV. Such a file can be
             # saved as a standard JPEG. RGB JPEGs without YUV conversion can't
             # be saved as JPEGs, and are probably bugs. Some software in the
             # wild actually produces RGB JPEGs in PDFs (probably a bug).
             DEFAULT_CT_RGB = 1
-            ct = self.filter_decodeparms[0][1].get('/ColorTransform', DEFAULT_CT_RGB)
+            ct = DEFAULT_CT_RGB
+            if self.filter_decodeparms[0][1] is not None:
+                ct = self.filter_decodeparms[0][1].get(
+                    '/ColorTransform', DEFAULT_CT_RGB
+                )
             return self.mode == 'RGB' and ct == DEFAULT_CT_RGB
 
-        def normal_dct_cmyk():
+        def normal_dct_cmyk() -> bool:
             # Normal DCTDecode CMYKs have /ColorTransform 0 and can be saved.
             # There is a YUVK colorspace but CMYK JPEGs don't generally use it
             DEFAULT_CT_CMYK = 0
-            ct = self.filter_decodeparms[0][1].get('/ColorTransform', DEFAULT_CT_CMYK)
+            ct = DEFAULT_CT_CMYK
+            if self.filter_decodeparms[0][1] is not None:
+                ct = self.filter_decodeparms[0][1].get(
+                    '/ColorTransform', DEFAULT_CT_CMYK
+                )
             return self.mode == 'CMYK' and ct == DEFAULT_CT_CMYK
 
         data, filters = self._remove_simple_filters(self.obj, self.filters)
@@ -509,7 +510,7 @@ class PdfImage(PdfImageBase):
             stream.write(self._generate_ccitt_header(data, icc=icc))
             stream.write(data)
             return '.tif'
-        elif filters == ['/DCTDecode'] and (
+        if filters == ['/DCTDecode'] and (
             self.mode == 'L' or normal_dct_rgb() or normal_dct_cmyk()
         ):
             stream.write(data)
@@ -517,19 +518,63 @@ class PdfImage(PdfImageBase):
 
         raise NotExtractableError()
 
-    def _extract_transcoded(self) -> Image.Image:
-        # Reminder Pillow palette byte order unintentionally changed in 8.3.0
-        # https://github.com/python-pillow/Pillow/issues/5595
-        # 8.2.0: all aligned by channel (very nonstandard)
-        # 8.3.0: all channels for one color followed by the next color (e.g. RGBRGBRGB)
+    def _extract_transcoded_1248bits(self) -> Image.Image:
+        """Extract an image when there are 1/2/4/8 bits packed in byte data."""
+        stride = 0  # tell Pillow to calculate stride from line width
+        scale = 0 if self.mode == 'L' else 1
+        if self.bits_per_component in (2, 4):
+            buffer, stride = _transcoding.unpack_subbyte_pixels(
+                self.read_bytes(), self.size, self.bits_per_component, scale
+            )
+        elif self.bits_per_component == 8:
+            buffer = cast(memoryview, self.get_stream_buffer())
+        else:
+            raise InvalidPdfImageError("BitsPerComponent must be 1, 2, 4, 8, or 16")
 
-        im = None
+        if self.mode == 'P' and self.palette is not None:
+            base_mode, palette = self.palette
+            im = _transcoding.image_from_buffer_and_palette(
+                buffer,
+                self.size,
+                stride,
+                base_mode,
+                palette,
+            )
+        else:
+            im = _transcoding.image_from_byte_buffer(buffer, self.size, stride)
+        return im
+
+    def _extract_transcoded_1bit(self) -> Image.Image:
+        if self.mode in ('RGB', 'CMYK'):
+            raise UnsupportedImageTypeError("1-bit RGB and CMYK are not supported")
+        try:
+            data = self.read_bytes()
+        except (RuntimeError, PdfError) as e:
+            if (
+                'read_bytes called on unfilterable stream' in str(e)
+                and not jbig2.get_decoder().available()
+            ):
+                raise DependencyError(
+                    "jbig2dec - not installed or installed version is too old "
+                    "(older than version 0.15)"
+                ) from None
+            raise
+
+        im = Image.frombytes('1', self.size, data)
+
+        if self.palette is not None:
+            base_mode, palette = self.palette
+            im = _transcoding.fix_1bit_palette_image(im, base_mode, palette)
+
+        return im
+
+    def _extract_transcoded(self) -> Image.Image:
         if self.mode in {'DeviceN', 'Separation'}:
             raise HifiPrintImageNotTranscodableError()
 
         if self.mode == 'RGB' and self.bits_per_component == 8:
-            # No point in accessing the buffer here, size qpdf decodes to 3-byte
-            # RGB and Pillow needs RGBX for raw access
+            # Cannot use the zero-copy .get_stream_buffer here, we have 3-byte
+            # RGB and Pillow needs RGBX.
             im = Image.frombuffer(
                 'RGB', self.size, self.read_bytes(), 'raw', 'RGB', 0, 1
             )
@@ -537,80 +582,21 @@ class PdfImage(PdfImageBase):
             im = Image.frombuffer(
                 'CMYK', self.size, self.get_stream_buffer(), 'raw', 'CMYK', 0, 1
             )
-        elif self.mode in ('L', 'P') and self.bits_per_component == 8:
-            buffer = self.get_stream_buffer()
-            stride = 0  # tell Pillow to calculate stride from line width
-            ystep = 1  # image is top to bottom in memory
-            im = Image.frombuffer('L', self.size, buffer, "raw", 'L', stride, ystep)
-            if self.mode == 'P' and self.palette is not None:
-                base_mode, palette = self.palette
-                if base_mode == 'RGB':
-                    im.putpalette(palette, rawmode=base_mode)
-                elif base_mode == 'L':
-                    # Pillow does not fully support palettes with rawmode='L'.
-                    # Convert to RGB palette.
-                    gray_palette = palette
-                    palette = b''
-                    for entry in gray_palette:
-                        palette += bytes([entry]) * 3
-                    im.putpalette(palette, rawmode='RGB')
-                elif base_mode == 'CMYK':
-                    # Pillow does not support CMYK with palettes; convert manually
-                    with memoryview(buffer) as mv:
-                        output = bytearray(4 * len(mv))
-                        for n, pal_idx in enumerate(mv):
-                            output[4 * n : 4 * (n + 1)] = palette[
-                                4 * pal_idx : 4 * (pal_idx + 1)
-                            ]
-                    im = Image.frombuffer(
-                        'CMYK', self.size, data=output, decoder_name='raw'
-                    )
-                else:
-                    raise NotImplementedError(f'palette with {base_mode}')
+        # elif self.mode == '1':
         elif self.bits_per_component == 1:
-            if self.filters and self.filters[0] == '/JBIG2Decode':
-                if not jbig2.jbig2dec_available():
-                    raise DependencyError(
-                        "jbig2dec - not installed or installed version is too old "
-                        "(older than version 0.15)"
-                    )
-                jbig2_globals_obj = self.filter_decodeparms[0][1].get('/JBIG2Globals')
-                im = jbig2.extract_jbig2(self.obj, jbig2_globals_obj)
-            else:
-                data = self.read_bytes()
-                im = Image.frombytes('1', self.size, data)
+            im = self._extract_transcoded_1bit()
+        elif self.mode in ('L', 'P') and self.bits_per_component <= 8:
+            im = self._extract_transcoded_1248bits()
         else:
             raise UnsupportedImageTypeError(repr(self) + ", " + repr(self.obj))
 
-        if (
-            self.mode == 'P'
-            and self.bits_per_component == 1
-            and self.palette is not None
-        ):
-            # Fix paletted 1-bit images
-            base_mode, palette = self.palette
-            if base_mode == 'RGB' and palette != b'\x00\x00\x00\xff\xff\xff':
-                im = im.convert('P')
-                im.putpalette(palette, rawmode=base_mode)
-                gp = im.getpalette()
-                if gp:
-                    gp[765:768] = gp[3:6]  # work around Pillow bug
-                    im.putpalette(gp)
-            elif base_mode == 'L' and palette != b'\x00\xff':
-                im = im.convert('P')
-                im.putpalette(palette, rawmode=base_mode)
-                gp = im.getpalette()
-                if gp:
-                    gp[255] = gp[1]  # work around Pillow bug
-                    im.putpalette(gp)
-
-        if self.colorspace == '/ICCBased':
+        if self.colorspace == '/ICCBased' and self.icc is not None:
             im.info['icc_profile'] = self.icc.tobytes()
 
         return im
 
-    def _extract_to_stream(self, *, stream) -> str:
-        """Attempt to extract the image to a stream
+    def _extract_to_stream(self, *, stream: BinaryIO) -> str:
+        """Extract the image to a stream.
 
         If possible, the compressed data is extracted and inserted into
         a compressed image file format without transcoding the compressed
@@ -623,7 +609,6 @@ class PdfImage(PdfImageBase):
         Returns:
             The file format extension.
         """
-
         try:
             return self._extract_direct(stream=stream)
         except NotExtractableError:
@@ -635,7 +620,7 @@ class PdfImage(PdfImageBase):
             if im.mode == 'CMYK':
                 im.save(stream, format='tiff', compression='tiff_adobe_deflate')
                 return '.tiff'
-            elif im:
+            if im:
                 im.save(stream, format='png')
                 return '.png'
         except PdfError as e:
@@ -648,8 +633,10 @@ class PdfImage(PdfImageBase):
 
         raise UnsupportedImageTypeError(repr(self))
 
-    def extract_to(self, *, stream=None, fileprefix='') -> str:
-        """Attempt to extract the image directly to a usable image file
+    def extract_to(
+        self, *, stream: BinaryIO | None = None, fileprefix: str = ''
+    ) -> str:
+        """Extract the image directly to a usable image file.
 
         If possible, the compressed data is extracted and inserted into
         a compressed image file format without transcoding the compressed
@@ -664,7 +651,6 @@ class PdfImage(PdfImageBase):
         Images might be saved as any of .png, .jpg, or .tiff.
 
         Examples:
-
             >>> im.extract_to(stream=bytes_io)
             '.png'
 
@@ -681,7 +667,6 @@ class PdfImage(PdfImageBase):
             appropriate extension. If no *fileprefix*, then an extension
             indicating the file type.
         """
-
         if bool(stream) == bool(fileprefix):
             raise ValueError("Cannot set both stream and fileprefix")
         if stream:
@@ -695,11 +680,15 @@ class PdfImage(PdfImageBase):
             copyfileobj(bio, target)
         return str(filepath)
 
-    def read_bytes(self, decode_level=StreamDecodeLevel.specialized):
+    def read_bytes(
+        self, decode_level: StreamDecodeLevel = StreamDecodeLevel.specialized
+    ) -> bytes:
         """Decompress this image and return it as unencoded bytes."""
         return self.obj.read_bytes(decode_level=decode_level)
 
-    def get_stream_buffer(self, decode_level=StreamDecodeLevel.specialized):
+    def get_stream_buffer(
+        self, decode_level: StreamDecodeLevel = StreamDecodeLevel.specialized
+    ) -> Buffer:
         """Access this image with the buffer protocol."""
         return self.obj.get_stream_buffer(decode_level=decode_level)
 
@@ -722,14 +711,13 @@ class PdfImage(PdfImageBase):
 
         return im
 
-    def _generate_ccitt_header(self, data, icc=None):
+    def _generate_ccitt_header(self, data: bytes, icc: bytes | None = None) -> bytes:
         """Construct a CCITT G3 or G4 header from the PDF metadata."""
         # https://stackoverflow.com/questions/2641770/
         # https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf
 
         if not self.decode_parms:
             raise ValueError("/CCITTFaxDecode without /DecodeParms")
-
         if self.decode_parms[0].get("/EncodedByteAlign", False):
             raise UnsupportedImageTypeError(
                 "/CCITTFaxDecode with /EncodedByteAlign true"
@@ -742,69 +730,27 @@ class PdfImage(PdfImageBase):
             ccitt_group = 3  # Group 3 2-D
         else:
             ccitt_group = 2  # Group 3 1-D
-        black_is_one = self.decode_parms[0].get("/BlackIs1", False)
+        _black_is_one = self.decode_parms[0].get("/BlackIs1", False)
+        # PDF spec says:
+        # BlackIs1: A flag indicating whether 1 bits shall be interpreted as black
+        # pixels and 0 bits as white pixels, the reverse of the normal
+        # PDF convention for image data. Default value: false.
         # TIFF spec says:
-        # use 0 for white_is_zero (=> black is 1)
-        # use 1 for black_is_zero (=> white is 1)
-        # In practice, do the opposite of what the TIFF spec says.
-        photometry = 1 if black_is_one else 0
+        # use 0 for white_is_zero (=> black is 1) MINISWHITE
+        # use 1 for black_is_zero (=> white is 1) MINISBLACK
+        # However, despite the documentation, it seems PDF viewers treat
+        # photometry as 0 when ccitt is involved.
+        # For example see
+        # https://gitlab.gnome.org/GNOME/evince/-/blob/main/backend/tiff/tiff2ps.c#L852-865
+        photometry = 0
 
         img_size = len(data)
-        tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H'
-
-        tag_keys = {tag.name: key for key, tag in TIFF_TAGS.items()}
-        ifd_struct = '<HHLL'
-
         if icc is None:
             icc = b''
-
-        ifds = []
-
-        def header_length(ifd_count):
-            return (
-                struct.calcsize(tiff_header_struct)
-                + struct.calcsize(ifd_struct) * ifd_count
-                + 4
-            )
-
-        def add_ifd(tag_name, data, count=1):
-            key = tag_keys[tag_name]
-            typecode = TIFF_TAGS[key].type
-            ifds.append((key, typecode, count, data))
-
-        image_offset = None
-        add_ifd('ImageWidth', self.width)
-        add_ifd('ImageLength', self.height)
-        add_ifd('BitsPerSample', 1)
-        add_ifd('Compression', ccitt_group)
-        add_ifd('PhotometricInterpretation', int(photometry))
-        add_ifd('StripOffsets', lambda: image_offset)
-        add_ifd('RowsPerStrip', self.height)
-        add_ifd('StripByteCounts', img_size)
-
-        icc_offset = 0
-        if icc:
-            add_ifd('ICCProfile', lambda: icc_offset, count=len(icc))
-
-        icc_offset = header_length(len(ifds))
-        image_offset = icc_offset + len(icc)
-
-        ifd_args = [(arg() if callable(arg) else arg) for ifd in ifds for arg in ifd]
-        tiff_header = struct.pack(
-            (tiff_header_struct + ifd_struct[1:] * len(ifds) + 'L'),
-            b'II',  # Byte order indication: Little endian
-            42,  # Version number (always 42)
-            8,  # Offset to first IFD
-            len(ifds),  # Number of tags in IFD
-            *ifd_args,
-            0,  # Last IFD
+        return _transcoding.generate_ccitt_header(
+            self.size, img_size, ccitt_group, photometry, icc
         )
 
-        if icc:
-            tiff_header += icc
-
-        return tiff_header
-
     def show(self):  # pragma: no cover
         """Show the image however PIL wants to."""
         self.as_pil_image().show()
@@ -815,7 +761,7 @@ class PdfImage(PdfImageBase):
             f'size={self.width}x{self.height} at {hex(id(self))}>'
         )
 
-    def _repr_png_(self):
+    def _repr_png_(self) -> bytes:
         """Display hook for IPython/Jupyter."""
         b = BytesIO()
         with self.as_pil_image() as im:
@@ -824,9 +770,14 @@ class PdfImage(PdfImageBase):
 
 
 class PdfJpxImage(PdfImage):
-    """Support class for JPEG 2000 images. Implements the same API as :class:`PdfImage`."""
+    """Support class for JPEG 2000 images. Implements the same API as :class:`PdfImage`.
+
+    If you call PdfImage(object_that_is_actually_jpeg2000_image), pikepdf will return
+    this class instead, due to the check in PdfImage.__new__.
+    """
 
     def __init__(self, obj):
+        """Initialize a JPEG 2000 image."""
         super().__init__(obj)
         self._jpxpil = self.as_pil_image()
 
@@ -839,7 +790,7 @@ class PdfJpxImage(PdfImage):
             and self._jpxpil == other._jpxpil
         )
 
-    def _extract_direct(self, *, stream):
+    def _extract_direct(self, *, stream: BinaryIO):
         data, filters = self._remove_simple_filters(self.obj, self.filters)
         if filters != ['/JPXDecode']:
             raise UnsupportedImageTypeError(self.filters)
@@ -848,6 +799,11 @@ class PdfJpxImage(PdfImage):
 
     @property
     def _colorspaces(self):
+        """Return the effective colorspace of a JPEG 2000 image.
+
+        If the ColorSpace dictionary is present, the colorspace embedded in the
+        JPEG 2000 data will be ignored, as required by the specification.
+        """
         # (PDF 1.7 Table 89) If ColorSpace is present, any colour space
         # specifications in the JPEG2000 data shall be ignored.
         super_colorspaces = super()._colorspaces
@@ -855,12 +811,13 @@ class PdfJpxImage(PdfImage):
             return super_colorspaces
         if self._jpxpil.mode == 'L':
             return ['/DeviceGray']
-        elif self._jpxpil.mode == 'RGB':
+        if self._jpxpil.mode == 'RGB':
             return ['/DeviceRGB']
         raise NotImplementedError('Complex JP2 colorspace')
 
     @property
-    def _bpc(self):
+    def _bpc(self) -> int:
+        """Return 8, since bpc is not meaningful for JPEG 2000 encoding."""
         # (PDF 1.7 Table 89) If the image stream uses the JPXDecode filter, this
         # entry is optional and shall be ignored if present. The bit depth is
         # determined by the conforming reader in the process of decoding the
@@ -868,7 +825,8 @@ class PdfJpxImage(PdfImage):
         return 8
 
     @property
-    def indexed(self):
+    def indexed(self) -> bool:
+        """Return False, since JPEG 2000 should not be indexed."""
         # Nothing in the spec precludes an Indexed JPXDecode image, except for
         # the fact that doing so is madness. Let's assume it no one is that
         # insane.
@@ -906,13 +864,16 @@ class PdfInlineImage(PdfImageBase):
     }
     REVERSE_ABBREVS = {v: k for k, v in ABBREVS.items()}
 
+    _data: Object
+    _image_object: tuple[Object, ...]
+
     def __init__(self, *, image_data: Object, image_object: tuple):
-        """
+        """Construct wrapper for inline image.
+
         Args:
             image_data: data stream for image, extracted from content stream
             image_object: the metadata for image, also from content stream
         """
-
         # Convert the sequence of pikepdf.Object from the content stream into
         # a dictionary object by unparsing it (to bytes), eliminating inline
         # image abbreviations, and constructing a bytes string equivalent to
@@ -949,19 +910,19 @@ class PdfInlineImage(PdfImageBase):
                 name = obj.unparse(resolved=True)
                 assert isinstance(name, bytes)
                 return remap_names.get(name, name)
-            else:
-                return obj.unparse(resolved=True)
-        elif isinstance(obj, bool):
+            return obj.unparse(resolved=True)
+        if isinstance(obj, bool):
             return b'true' if obj else b'false'  # Lower case for PDF spec
-        elif isinstance(obj, (int, Decimal, float)):
+        if isinstance(obj, (int, Decimal, float)):
             return str(obj).encode('ascii')
-        else:
-            raise NotImplementedError(repr(obj))
+        raise NotImplementedError(repr(obj))
 
     def _metadata(self, name, type_, default):
-        return metadata_from_obj(self.obj, name, type_, default)
+        return _metadata_from_obj(self.obj, name, type_, default)
+
+    def unparse(self) -> bytes:
+        """Create the content stream bytes that reproduce this inline image."""
 
-    def unparse(self):
         def metadata_tokens():
             for metadata_obj in self._image_object:
                 unparsed = self._unparse_obj(
@@ -979,12 +940,9 @@ class PdfInlineImage(PdfImageBase):
 
         return b''.join(inline_image_tokens())
 
-    @property
-    def is_inline(self):
-        return True
-
     @property
     def icc(self):  # pragma: no cover
+        """Raise an exception since ICC profiles are not supported on inline images."""
         raise InvalidPdfImageError(
             "Inline images with ICC profiles are not supported in the PDF specification"
         )
@@ -1016,18 +974,26 @@ class PdfInlineImage(PdfImageBase):
         img = PdfImage(raw_img)
         return img
 
-    def as_pil_image(self):
+    def as_pil_image(self) -> Image.Image:
+        """Return inline image as a Pillow Image."""
         return self._convert_to_pdfimage().as_pil_image()
 
-    def extract_to(self, *, stream=None, fileprefix=''):
+    def extract_to(self, *, stream: BinaryIO | None = None, fileprefix: str = ''):
+        """Extract the inline image directly to a usable image file.
+
+        See:
+            :meth:`PdfImage.extract_to`
+        """
         return self._convert_to_pdfimage().extract_to(
             stream=stream, fileprefix=fileprefix
         )
 
     def read_bytes(self):
-        raise NotImplementedError("qpdf returns compressed")
-        # return self._data._inline_image_bytes()
+        """Return decompressed image bytes."""
+        # QPDF does not have an API to return this directly, so convert it.
+        return self._convert_to_pdfimage().read_bytes()
 
     def get_stream_buffer(self):
-        raise NotImplementedError("qpdf returns compressed")
-        # return memoryview(self._data.inline_image_bytes())
+        """Return decompressed stream buffer."""
+        # QPDF does not have an API to return this directly, so convert it.
+        return self._convert_to_pdfimage().get_stream_buffer()
