@@ -1,28 +1,18 @@
-# This Source Code Form is subject to the terms of the Mozilla Public
-# License, v. 2.0. If a copy of the MPL was not distributed with this
-# file, You can obtain one at http://mozilla.org/MPL/2.0/.
-#
-# Copyright (C) 2018, James R. Barlow (https://github.com/jbarlow83/)
+# SPDX-FileCopyrightText: 2022 James R. Barlow
+# SPDX-License-Identifier: MPL-2.0
+
+"""PDF metadata handling."""
+
+from __future__ import annotations
 
 import logging
 import re
 import sys
-from abc import ABC, abstractstaticmethod
+from abc import ABC, abstractmethod
 from datetime import datetime
 from functools import wraps
 from io import BytesIO
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    Dict,
-    List,
-    NamedTuple,
-    Optional,
-    Set,
-    Type,
-    Union,
-)
+from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Set
 from warnings import warn
 
 from lxml import etree
@@ -40,11 +30,13 @@ else:
 if TYPE_CHECKING:  # pragma: no cover
     from pikepdf import Pdf
 
+
 XMP_NS_DC = "http://purl.org/dc/elements/1.1/"
 XMP_NS_PDF = "http://ns.adobe.com/pdf/1.3/"
 XMP_NS_PDFA_ID = "http://www.aiim.org/pdfa/ns/id/"
 XMP_NS_PDFX_ID = "http://www.npes.org/pdfx/ns/id/"
 XMP_NS_PHOTOSHOP = "http://ns.adobe.com/photoshop/1.0/"
+XMP_NS_PRISM = "http://prismstandard.org/namespaces/basic/1.0/"
 XMP_NS_PRISM2 = "http://prismstandard.org/namespaces/basic/2.0/"
 XMP_NS_PRISM3 = "http://prismstandard.org/namespaces/basic/3.0/"
 XMP_NS_RDF = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
@@ -52,13 +44,14 @@ XMP_NS_XMP = "http://ns.adobe.com/xap/1.0/"
 XMP_NS_XMP_MM = "http://ns.adobe.com/xap/1.0/mm/"
 XMP_NS_XMP_RIGHTS = "http://ns.adobe.com/xap/1.0/rights/"
 
-DEFAULT_NAMESPACES = [
+DEFAULT_NAMESPACES: list[tuple[str, str]] = [
     ('adobe:ns:meta/', 'x'),
     (XMP_NS_DC, 'dc'),
     (XMP_NS_PDF, 'pdf'),
     (XMP_NS_PDFA_ID, 'pdfaid'),
     (XMP_NS_PDFX_ID, 'pdfxid'),
     (XMP_NS_PHOTOSHOP, 'photoshop'),
+    (XMP_NS_PRISM, 'prism'),
     (XMP_NS_PRISM2, 'prism2'),
     (XMP_NS_PRISM3, 'prism3'),
     (XMP_NS_RDF, 'rdf'),
@@ -85,22 +78,22 @@ XPACKET_END = b"""\n<?xpacket end="w"?>\n"""
 
 
 class XmpContainer(NamedTuple):
+    """Map XMP container object to suitable Python container."""
+
     rdf_type: str
-    py_type: Type
-    insert_fn: Callable
+    py_type: type
+    insert_fn: Callable[..., None]
 
 
 log = logging.getLogger(__name__)
 
 
 class NeverRaise(Exception):
-    """An exception that is never raised"""
-
-    pass  # pylint: disable=unnecessary-pass
+    """An exception that is never raised."""
 
 
 class AltList(list):
-    pass
+    """XMP AltList container."""
 
 
 XMP_CONTAINERS = [
@@ -129,24 +122,44 @@ re_xml_illegal_bytes = re.compile(
 )
 
 
-def _clean(s: Union[str, Iterable[str]], joiner: str = '; ') -> str:
+def _parser_basic(xml: bytes):
+    return parse_xml(BytesIO(xml))
+
+
+def _parser_strip_illegal_bytes(xml: bytes):
+    return parse_xml(BytesIO(re_xml_illegal_bytes.sub(b'', xml)))
+
+
+def _parser_recovery(xml: bytes):
+    return parse_xml(BytesIO(xml), recover=True)
+
+
+def _parser_replace_with_empty_xmp(_xml: bytes = b''):
+    log.warning("Error occurred parsing XMP, replacing with empty XMP.")
+    return _parser_basic(XMP_EMPTY)
+
+
+def _clean(s: str | Iterable[str], joiner: str = '; ') -> str:
     """Ensure an object can safely be inserted in a XML tag body.
 
     If we still have a non-str object at this point, the best option is to
     join it, because it's apparently calling for a new node in a place that
     isn't allowed in the spec or not supported.
     """
-    if not isinstance(s, str) and isinstance(s, Iterable):
-        warn(f"Merging elements of {s}")
-        if isinstance(s, Set):
-            s = joiner.join(sorted(s))
+    if not isinstance(s, str):
+        if isinstance(s, Iterable):
+            warn(f"Merging elements of {s}")
+            if isinstance(s, Set):
+                s = joiner.join(sorted(s))
+            else:
+                s = joiner.join(s)
         else:
-            s = joiner.join(s)
+            raise TypeError("object must be a string or iterable of strings")
     return re_xml_illegal_chars.sub('', s)
 
 
 def encode_pdf_date(d: datetime) -> str:
-    """Encode Python datetime object as PDF date string
+    """Encode Python datetime object as PDF date string.
 
     From Adobe pdfmark manual:
     (D:YYYYMMDDHHmmSSOHH'mm')
@@ -165,7 +178,6 @@ def encode_pdf_date(d: datetime) -> str:
 
     'D:' is required in PDF/A, so we always add it.
     """
-
     # The formatting of %Y is not consistent as described in
     # https://bugs.python.org/issue13305 and underspecification in libc.
     # So explicitly format the year with leading zeros
@@ -179,7 +191,7 @@ def encode_pdf_date(d: datetime) -> str:
 
 
 def decode_pdf_date(s: str) -> datetime:
-    """Decode a pdfmark date to a Python datetime object
+    """Decode a pdfmark date to a Python datetime object.
 
     A pdfmark date is a string in a paritcular format. See the pdfmark
     Reference for the specification.
@@ -203,84 +215,79 @@ def decode_pdf_date(s: str) -> datetime:
 
 
 class Converter(ABC):
-    @abstractstaticmethod
-    def xmp_from_docinfo(docinfo_val: Optional[str]) -> Any:
-        "Derive XMP metadata from a DocumentInfo string"
+    """XMP <-> DocumentInfo converter."""
 
-    @abstractstaticmethod
-    def docinfo_from_xmp(xmp_val: Any) -> Optional[str]:
-        "Derive a DocumentInfo value from equivalent XMP metadata"
+    @staticmethod
+    @abstractmethod
+    def xmp_from_docinfo(docinfo_val: str | None) -> Any:  # type: ignore
+        """Derive XMP metadata from a DocumentInfo string."""
+
+    @staticmethod
+    @abstractmethod
+    def docinfo_from_xmp(xmp_val: Any) -> str | None:
+        """Derive a DocumentInfo value from equivalent XMP metadata."""
 
 
 class AuthorConverter(Converter):
+    """Convert XMP document authors to DocumentInfo."""
+
     @staticmethod
-    def xmp_from_docinfo(docinfo_val):
+    def xmp_from_docinfo(docinfo_val: str | None) -> Any:  # type: ignore
+        """Derive XMP authors info from DocumentInfo."""
         return [docinfo_val]
 
     @staticmethod
     def docinfo_from_xmp(xmp_val):
+        """Derive DocumentInfo authors from XMP.
+
+        XMP supports multiple author values, while DocumentInfo has a string,
+        so we return the values separated by semi-colons.
+        """
         if isinstance(xmp_val, str):
             return xmp_val
-        elif xmp_val is None or xmp_val == [None]:
+        if xmp_val is None or xmp_val == [None]:
             return None
-        else:
-            return '; '.join(xmp_val)
-
-
-def _fromisoformat_py36(datestr: str) -> datetime:
-    """Backported equivalent of datetime.fromisoformat
-
-    Can remove whenever we drop Python 3.6 support.
-    """
-    # strptime %z can't parse a timezone with punctuation
-    if re.search(r'[+-]\d{2}[-:]\d{2}$', datestr):
-        datestr = datestr[:-3] + datestr[-2:]
-    formats = [
-        "%Y-%m-%dT%H:%M:%S%z",
-        "%Y-%m-%dT%H:%M:%S",
-        "%Y-%m-%dT%H:%M:%S.%f%z",
-        "%Y-%m-%dT%H:%M:%S.%f",
-    ]
-    for fmt in formats:
-        try:
-            return datetime.strptime(datestr, fmt)
-        except ValueError:
-            continue
-    raise ValueError(
-        f"Could not parse ISO date: {datestr}. Try Python 3.7+, which "
-        "has better support for ISO dates."
-    )
-
-
-if sys.version_info < (3, 7) or sys.implementation.name == 'pypy':
-    fromisoformat = _fromisoformat_py36
-else:
-    fromisoformat = datetime.fromisoformat
+        return '; '.join(xmp_val)
 
 
 class DateConverter(Converter):
+    """Convert XMP dates to DocumentInfo."""
+
     @staticmethod
     def xmp_from_docinfo(docinfo_val):
+        """Derive XMP date from DocumentInfo."""
         if docinfo_val == '':
             return ''
         return decode_pdf_date(docinfo_val).isoformat()
 
     @staticmethod
     def docinfo_from_xmp(xmp_val):
+        """Derive DocumentInfo from XMP."""
         if xmp_val.endswith('Z'):
             xmp_val = xmp_val[:-1] + '+00:00'
-        dateobj = fromisoformat(xmp_val)
+        try:
+            dateobj = datetime.fromisoformat(xmp_val)
+        except IndexError:
+            # PyPy 3.7 may raise IndexError - convert to ValueError
+            raise ValueError(f"Invalid isoformat string: '{xmp_val}'") from None
         return encode_pdf_date(dateobj)
 
 
 class DocinfoMapping(NamedTuple):
+    """Map DocumentInfo keys to their XMP equivalents, along with converter."""
+
     ns: str
     key: str
     name: Name
-    converter: Optional[Type[Converter]]
+    converter: type[Converter] | None
 
 
 def ensure_loaded(fn):
+    """Ensure the XMP has been loaded and parsed.
+
+    TODO: Can this be removed? Why allow the uninit'ed state to even exist?
+    """
+
     @wraps(fn)
     def wrapper(self, *args, **kwargs):
         if not self._xmp:
@@ -291,7 +298,7 @@ def ensure_loaded(fn):
 
 
 class PdfMetadata(MutableMapping):
-    """Read and edit the metadata associated with a PDF
+    """Read and edit the metadata associated with a PDF.
 
     The PDF specification contain two types of metadata, the newer XMP
     (Extensible Metadata Platform, XML-based) and older DocumentInformation
@@ -321,7 +328,7 @@ class PdfMetadata(MutableMapping):
         :meth:`pikepdf.Pdf.open_metadata`
     """
 
-    DOCINFO_MAPPING: List[DocinfoMapping] = [
+    DOCINFO_MAPPING: list[DocinfoMapping] = [
         DocinfoMapping(XMP_NS_DC, 'creator', Name.Author, AuthorConverter),
         DocinfoMapping(XMP_NS_DC, 'description', Name.Subject, None),
         DocinfoMapping(XMP_NS_DC, 'title', Name.Title, None),
@@ -332,12 +339,20 @@ class PdfMetadata(MutableMapping):
         DocinfoMapping(XMP_NS_XMP, 'ModifyDate', Name.ModDate, DateConverter),
     ]
 
-    NS: Dict[str, str] = {prefix: uri for uri, prefix in DEFAULT_NAMESPACES}
-    REVERSE_NS: Dict[str, str] = {uri: prefix for uri, prefix in DEFAULT_NAMESPACES}
+    NS: dict[str, str] = {prefix: uri for uri, prefix in DEFAULT_NAMESPACES}
+    REVERSE_NS: dict[str, str] = dict(DEFAULT_NAMESPACES)
+
+    _PARSERS_OVERWRITE_INVALID_XML: Iterable[Callable[[bytes], Any]] = [
+        _parser_basic,
+        _parser_strip_illegal_bytes,
+        _parser_recovery,
+        _parser_replace_with_empty_xmp,
+    ]
+    _PARSERS_STANDARD: Iterable[Callable[[bytes], Any]] = [_parser_basic]
 
     def __init__(
         self,
-        pdf: 'Pdf',
+        pdf: Pdf,
         pikepdf_mark: bool = True,
         sync_docinfo: bool = True,
         overwrite_invalid_xml: bool = True,
@@ -352,7 +367,7 @@ class PdfMetadata(MutableMapping):
     def load_from_docinfo(
         self, docinfo, delete_missing: bool = False, raise_failure: bool = False
     ) -> None:
-        """Populate the XMP metadata object with DocumentInfo
+        """Populate the XMP metadata object with DocumentInfo.
 
         Arguments:
             docinfo: a DocumentInfo, e.g pdf.docinfo
@@ -369,8 +384,7 @@ class PdfMetadata(MutableMapping):
         def warn_or_raise(msg, e=None):
             if raise_failure:
                 raise ValueError(msg) from e
-            else:
-                warn(msg)
+            warn(msg)
 
         for uri, shortkey, docinfo_name, converter in self.DOCINFO_MAPPING:
             qname = QName(uri, shortkey)
@@ -392,10 +406,10 @@ class PdfMetadata(MutableMapping):
                 warn_or_raise(
                     f"The metadata field {docinfo_name} could not be copied to XMP", e
                 )
-        valid_docinfo_names = set(
+        valid_docinfo_names = {
             str(docinfo_name) for _, _, docinfo_name, _ in self.DOCINFO_MAPPING
-        )
-        extra_docinfo_names = set(str(k) for k in docinfo.keys()) - valid_docinfo_names
+        }
+        extra_docinfo_names = {str(k) for k in docinfo.keys()} - valid_docinfo_names
         for extra in extra_docinfo_names:
             warn_or_raise(
                 f"The metadata field {extra} with value '{repr(docinfo.get(extra))}' "
@@ -406,57 +420,51 @@ class PdfMetadata(MutableMapping):
         try:
             data = self._pdf.Root.Metadata.read_bytes()
         except AttributeError:
-            data = XMP_EMPTY
+            data = b''
         self._load_from(data)
 
     def _load_from(self, data: bytes) -> None:
         if data.strip() == b'':
             data = XMP_EMPTY  # on some platforms lxml chokes on empty documents
 
-        def basic_parser(xml):
-            return parse_xml(BytesIO(xml))
-
-        def strip_illegal_bytes_parser(xml):
-            return parse_xml(BytesIO(re_xml_illegal_bytes.sub(b'', xml)))
-
-        def recovery_parser(xml):
-            return parse_xml(BytesIO(xml), recover=True)
-
-        def replace_with_empty_xmp(_xml=None):
-            log.warning("Error occurred parsing XMP, replacing with empty XMP.")
-            return basic_parser(XMP_EMPTY)
-
-        if self.overwrite_invalid_xml:
-            parsers: Iterable[Callable] = [
-                basic_parser,
-                strip_illegal_bytes_parser,
-                recovery_parser,
-                replace_with_empty_xmp,
-            ]
-        else:
-            parsers = [basic_parser]
+        parsers = (
+            self._PARSERS_OVERWRITE_INVALID_XML
+            if self.overwrite_invalid_xml
+            else self._PARSERS_STANDARD
+        )
 
         for parser in parsers:
             try:
                 self._xmp = parser(data)
-            except (XMLSyntaxError if self.overwrite_invalid_xml else NeverRaise) as e:
+            except (
+                XMLSyntaxError
+                if self.overwrite_invalid_xml
+                else NeverRaise  # type: ignore
+            ) as e:
                 if str(e).startswith("Start tag expected, '<' not found") or str(
                     e
                 ).startswith("Document is empty"):
-                    self._xmp = replace_with_empty_xmp()
+                    self._xmp = _parser_replace_with_empty_xmp()
                     break
             else:
                 break
 
-        try:
-            pis = self._xmp.xpath('/processing-instruction()')
-            for pi in pis:
-                etree.strip_tags(self._xmp, pi.tag)
-            self._get_rdf_root()
-        except (Exception if self.overwrite_invalid_xml else NeverRaise) as e:
-            log.warning("Error occurred parsing XMP", exc_info=e)
-            self._xmp = replace_with_empty_xmp()
-        return
+        if self._xmp is not None:
+            try:
+                pis = self._xmp.xpath('/processing-instruction()')
+                for pi in pis:
+                    etree.strip_tags(self._xmp, pi.tag)
+                self._get_rdf_root()
+            except (
+                Exception  # pylint: disable=broad-except
+                if self.overwrite_invalid_xml
+                else NeverRaise
+            ) as e:
+                log.warning("Error occurred parsing XMP", exc_info=e)
+                self._xmp = _parser_replace_with_empty_xmp()
+        else:
+            log.warning("Error occurred parsing XMP")
+            self._xmp = _parser_replace_with_empty_xmp()
 
     @ensure_loaded
     def __enter__(self):
@@ -472,7 +480,7 @@ class PdfMetadata(MutableMapping):
             self._updating = False
 
     def _update_docinfo(self):
-        """Update the PDF's DocumentInfo dictionary to match XMP metadata
+        """Update the PDF's DocumentInfo dictionary to match XMP metadata.
 
         The standard mapping is described here:
             https://www.pdfa.org/pdfa-metadata-xmp-rdf-dublin-core/
@@ -525,7 +533,7 @@ class PdfMetadata(MutableMapping):
         return xml_bytes
 
     def _apply_changes(self):
-        """Serialize our changes back to the PDF in memory
+        """Serialize our changes back to the PDF in memory.
 
         Depending how we are initialized, leave our metadata mark and producer.
         """
@@ -549,8 +557,8 @@ class PdfMetadata(MutableMapping):
             self._update_docinfo()
 
     @classmethod
-    def _qname(cls, name: Union[QName, str]) -> str:
-        """Convert name to an XML QName
+    def _qname(cls, name: QName | str) -> str:
+        """Convert name to an XML QName.
 
         e.g. pdf:Producer -> {http://ns.adobe.com/pdf/1.3/}Producer
         """
@@ -562,12 +570,19 @@ class PdfMetadata(MutableMapping):
             return name
         if name.startswith('{'):
             return name
-        prefix, tag = name.split(':', maxsplit=1)
+        try:
+            prefix, tag = name.split(':', maxsplit=1)
+        except ValueError:
+            # If missing the namespace, put it in the top level namespace
+            # To do this completely correct we actually need to figure out
+            # the namespace based on context defined by parent tags. That
+            #   https://www.w3.org/2001/tag/doc/qnameids.html
+            prefix, tag = 'x', name
         uri = cls.NS[prefix]
         return str(QName(uri, tag))
 
     def _prefix_from_uri(self, uriname):
-        """Given a fully qualified XML name, find a prefix
+        """Given a fully qualified XML name, find a prefix.
 
         e.g. {http://ns.adobe.com/pdf/1.3/}Producer -> pdf:Producer
         """
@@ -576,7 +591,7 @@ class PdfMetadata(MutableMapping):
         return self.REVERSE_NS[uri] + ':' + tag
 
     def _get_subelements(self, node):
-        """Gather the sub-elements attached to a node
+        """Gather the sub-elements attached to a node.
 
         Gather rdf:Bag and and rdf:Seq into set and list respectively. For
         alternate languages values, take the first language only for
@@ -607,8 +622,8 @@ class PdfMetadata(MutableMapping):
                 raise ValueError("Metadata seems to be XML but not XMP")
         return rdf
 
-    def _get_elements(self, name: Union[str, QName] = ''):
-        """Get elements from XMP
+    def _get_elements(self, name: str | QName = ''):
+        """Get elements from XMP.
 
         Core routine to find elements matching name within the XMP and yield
         them.
@@ -651,11 +666,11 @@ class PdfMetadata(MutableMapping):
         yield from (v[2] for v in self._get_elements(name))
 
     @ensure_loaded
-    def __contains__(self, key: Union[str, QName]):
+    def __contains__(self, key: str | QName):
         return any(self._get_element_values(key))
 
     @ensure_loaded
-    def __getitem__(self, key: Union[str, QName]):
+    def __getitem__(self, key: str | QName):
         try:
             return next(self._get_element_values(key))
         except StopIteration:
@@ -675,14 +690,24 @@ class PdfMetadata(MutableMapping):
 
     def _setitem(
         self,
-        key: Union[str, QName],
-        val: Union[Set[str], List[str], str],
+        key: str | QName,
+        val: set[str] | list[str] | str,
         applying_mark: bool = False,
     ):
         if not self._updating:
             raise RuntimeError("Metadata not opened for editing, use with block")
 
         qkey = self._qname(key)
+        self._setitem_check_args(key, val, applying_mark, qkey)
+
+        try:
+            # Update existing node
+            self._setitem_update(key, val, qkey)
+        except StopIteration:
+            # Insert a new node
+            self._setitem_insert(key, val)
+
+    def _setitem_check_args(self, key, val, applying_mark: bool, qkey: str) -> None:
         if (
             self.mark
             and not applying_mark
@@ -702,80 +727,76 @@ class PdfMetadata(MutableMapping):
         if isinstance(val, str) and qkey in (self._qname('dc:creator')):
             log.error(f"{key} should be set to a list of strings")
 
-        def add_array(node, items: Iterable):
-            rdf_type = next(
-                c.rdf_type for c in XMP_CONTAINERS if isinstance(items, c.py_type)
-            )
-            seq = etree.SubElement(node, str(QName(XMP_NS_RDF, rdf_type)))
-            tag_attrib: Optional[Dict[str, str]] = None
-            if rdf_type == 'Alt':
-                tag_attrib = {str(QName(XMP_NS_XML, 'lang')): 'x-default'}
-            for item in items:
-                el = etree.SubElement(
-                    seq, str(QName(XMP_NS_RDF, 'li')), attrib=tag_attrib
-                )
-                el.text = _clean(item)
+    def _setitem_add_array(self, node, items: Iterable) -> None:
+        rdf_type = next(
+            c.rdf_type for c in XMP_CONTAINERS if isinstance(items, c.py_type)
+        )
+        seq = etree.SubElement(node, str(QName(XMP_NS_RDF, rdf_type)))
+        tag_attrib: dict[str, str] | None = None
+        if rdf_type == 'Alt':
+            tag_attrib = {str(QName(XMP_NS_XML, 'lang')): 'x-default'}
+        for item in items:
+            el = etree.SubElement(seq, str(QName(XMP_NS_RDF, 'li')), attrib=tag_attrib)
+            el.text = _clean(item)
 
-        try:
-            # Locate existing node to replace
-            node, attrib, _oldval, _parent = next(self._get_elements(key))
-            if attrib:
-                if not isinstance(val, str):
-                    if qkey == self._qname('dc:creator'):
-                        # dc:creator incorrectly created as an attribute - we're
-                        # replacing it anyway, so remove the old one
-                        del node.attrib[qkey]
-                        add_array(node, _clean(val))
-                    else:
-                        raise TypeError(f"Setting {key} to {val} with type {type(val)}")
+    def _setitem_update(self, key, val, qkey):
+        # Locate existing node to replace
+        node, attrib, _oldval, _parent = next(self._get_elements(key))
+        if attrib:
+            if not isinstance(val, str):
+                if qkey == self._qname('dc:creator'):
+                    # dc:creator incorrectly created as an attribute - we're
+                    # replacing it anyway, so remove the old one
+                    del node.attrib[qkey]
+                    self._setitem_add_array(node, _clean(val))
                 else:
-                    node.set(attrib, _clean(val))
-            elif isinstance(val, (list, set)):
-                for child in node.findall('*'):
-                    node.remove(child)
-                add_array(node, val)
-            elif isinstance(val, str):
-                for child in node.findall('*'):
-                    node.remove(child)
-                if str(self._qname(key)) in LANG_ALTS:
-                    add_array(node, AltList([_clean(val)]))
-                else:
-                    node.text = _clean(val)
+                    raise TypeError(f"Setting {key} to {val} with type {type(val)}")
             else:
-                raise TypeError(f"Setting {key} to {val} with type {type(val)}")
-        except StopIteration:
-            # Insert a new node
-            rdf = self._get_rdf_root()
+                node.set(attrib, _clean(val))
+        elif isinstance(val, (list, set)):
+            for child in node.findall('*'):
+                node.remove(child)
+            self._setitem_add_array(node, val)
+        elif isinstance(val, str):
+            for child in node.findall('*'):
+                node.remove(child)
             if str(self._qname(key)) in LANG_ALTS:
-                val = AltList([_clean(val)])
-            if isinstance(val, (list, set)):
-                rdfdesc = etree.SubElement(
-                    rdf,
-                    str(QName(XMP_NS_RDF, 'Description')),
-                    attrib={str(QName(XMP_NS_RDF, 'about')): ''},
-                )
-                node = etree.SubElement(rdfdesc, self._qname(key))
-                add_array(node, val)
-            elif isinstance(val, str):
-                _rdfdesc = etree.SubElement(  # lgtm [py/unused-local-variable]
-                    rdf,
-                    str(QName(XMP_NS_RDF, 'Description')),
-                    attrib={
-                        QName(XMP_NS_RDF, 'about'): '',
-                        self._qname(key): _clean(val),
-                    },
-                )
+                self._setitem_add_array(node, AltList([_clean(val)]))
             else:
-                raise TypeError(
-                    f"Setting {key} to {val} with type {type(val)}"
-                ) from None
+                node.text = _clean(val)
+        else:
+            raise TypeError(f"Setting {key} to {val} with type {type(val)}")
+
+    def _setitem_insert(self, key, val):
+        rdf = self._get_rdf_root()
+        if str(self._qname(key)) in LANG_ALTS:
+            val = AltList([_clean(val)])
+        if isinstance(val, (list, set)):
+            rdfdesc = etree.SubElement(
+                rdf,
+                str(QName(XMP_NS_RDF, 'Description')),
+                attrib={str(QName(XMP_NS_RDF, 'about')): ''},
+            )
+            node = etree.SubElement(rdfdesc, self._qname(key))
+            self._setitem_add_array(node, val)
+        elif isinstance(val, str):
+            _rdfdesc = etree.SubElement(
+                rdf,
+                str(QName(XMP_NS_RDF, 'Description')),
+                attrib={
+                    QName(XMP_NS_RDF, 'about'): '',
+                    self._qname(key): _clean(val),
+                },
+            )
+        else:
+            raise TypeError(f"Setting {key} to {val} with type {type(val)}") from None
 
     @ensure_loaded
-    def __setitem__(self, key: Union[str, QName], val: Union[Set[str], List[str], str]):
+    def __setitem__(self, key: str | QName, val: set[str] | list[str] | str):
         return self._setitem(key, val, False)
 
     @ensure_loaded
-    def __delitem__(self, key: Union[str, QName]):
+    def __delitem__(self, key: str | QName):
         if not self._updating:
             raise RuntimeError("Metadata not opened for editing, use with block")
         try:
@@ -795,19 +816,22 @@ class PdfMetadata(MutableMapping):
             raise KeyError(key) from None
 
     @property
-    @ensure_loaded
     def pdfa_status(self) -> str:
-        """Returns the PDF/A conformance level claimed by this PDF, or False
+        """Return the PDF/A conformance level claimed by this PDF, or False.
 
         A PDF may claim to PDF/A compliant without this being true. Use an
         independent verifier such as veraPDF to test if a PDF is truly
         conformant.
 
         Returns:
-            str: The conformance level of the PDF/A, or an empty string if the
+            The conformance level of the PDF/A, or an empty string if the
             PDF does not claim PDF/A conformance. Possible valid values
             are: 1A, 1B, 2A, 2B, 2U, 3A, 3B, 3U.
         """
+        # do same as @ensure_loaded - mypy can't handle decorated property
+        if not self._xmp:
+            self._load()
+
         key_part = QName(XMP_NS_PDFA_ID, 'part')
         key_conformance = QName(XMP_NS_PDFA_ID, 'conformance')
         try:
@@ -816,18 +840,21 @@ class PdfMetadata(MutableMapping):
             return ''
 
     @property
-    @ensure_loaded
     def pdfx_status(self) -> str:
-        """Returns the PDF/X conformance level claimed by this PDF, or False
+        """Return the PDF/X conformance level claimed by this PDF, or False.
 
         A PDF may claim to PDF/X compliant without this being true. Use an
         independent verifier such as veraPDF to test if a PDF is truly
         conformant.
 
         Returns:
-            str: The conformance level of the PDF/X, or an empty string if the
+            The conformance level of the PDF/X, or an empty string if the
             PDF does not claim PDF/X conformance.
         """
+        # do same as @ensure_loaded - mypy can't handle decorated property
+        if not self._xmp:
+            self._load()
+
         pdfx_version = QName(XMP_NS_PDFX_ID, 'GTS_PDFXVersion')
         try:
             return self[pdfx_version]
